//silent driver with no sleep and chooseable samplerate
#define VC_EXTRALEAN

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <process.h>
#include "../AudioDriver.h"

#define DEFAULT_SAMPLERATE	44100
#define BUFFERSIZE			16384*128		// size of virtual dma buffer

HINSTANCE dllInstance;
BOOL WINAPI DllMain ( HANDLE hModule, DWORD fwdreason, LPVOID lpReserved )
{
	switch (fwdreason) 
	{
	case DLL_PROCESS_ATTACH:
		dllInstance = (HINSTANCE) hModule;  
		break;
	case DLL_THREAD_ATTACH: 
		break;
	case DLL_THREAD_DETACH: 
		break;
	case DLL_PROCESS_DETACH: 
		break;
	}
	return TRUE;
}

class ad : public CAudioDriver
{
public:
	ad();
	virtual ~ad();
	virtual void Initialize(dword hwnd, float *(*pcallback)(int &numsamples));
	virtual void Reset();
	virtual bool Enable(bool e);	
	virtual int GetWritePos();
	virtual int GetPlayPos();
	virtual void Configure();
	virtual void EnableConfigDialog();

	void DoBlocks();

private:
	bool Start();	
	bool Stop();
	void ReadConfig();
	void Error(char const *msg);

	int GetTime();

public:
	double PerfFreq;
	double StartTime;
	float *(*pCallback)(int &numsamples);
	int WritePos;
	bool Initialized;
	bool Running;
	bool StopPolling;

	HWND LB_SAMPLERATE;
	HWND BUZZWND;
	void WriteConfig();
};

static void __cdecl PollerThread(void *poo)
{
	ad *pad = (ad *)poo;

	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);

	pad->StopPolling = false;

	while(!pad->StopPolling)
	{
		pad->DoBlocks();
		Sleep(0); //was 20
	}

	pad->StopPolling = false;

	_endthread();
}

ad::ad()
{
	Initialized = false;
}

ad::~ad()
{
	Reset();
}

void ad::Error(char const *msg)
{
	MessageBox(NULL, msg, "Error in Waveout Driver", MB_OK);
}

bool ad::Start()
{
	if (Running)
	{
		return true;
	}

	//LARGE_INTEGER li;
	//QueryPerformanceFrequency(&li);
	//PerfFreq = (double)li.LowPart + ((double)li.HighPart * (65536.0*65536.0));

	//QueryPerformanceCounter(&li);
	//StartTime = (double)li.LowPart + ((double)li.HighPart * (65536.0*65536.0));

	//WritePos = 0;

	HANDLE h = (HANDLE)_beginthread(PollerThread, 0, this);

	Running = true;
	return true;
}

bool ad::Stop()
{
	if (!Running)
	{
		return true;
	}

	StopPolling = true;
	while(StopPolling)
	{
		Sleep(20);
	}

	Running = false;
	return true;
}

void ad::DoBlocks()
{
	int n = BUFFERSIZE;
	pCallback(n);
} 

void ad::ReadConfig()
{
	SamplesPerSec = ReadProfileInt("SamplesPerSec", DEFAULT_SAMPLERATE);
	Flags = ADF_STEREO;
}

void ad::WriteConfig()
{
	WriteProfileInt("SamplesPerSec", SamplesPerSec);
}

void ad::Initialize(dword hwnd, float *(*pcallback)(int &numsamples))
{
	pCallback = pcallback;
	Running = false;
	ReadConfig();

	Initialized = true;
	Start();
}

void ad::Reset()
{
	Stop();
}

LRESULT CALLBACK SettingsDialog(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) 
{
	ad *pAD = (ad *)GetWindowLong(hwnd, GWL_USERDATA);

	switch (iMsg) 
	{
	case WM_CREATE :
		{
			CREATESTRUCT *cs = (LPCREATESTRUCT)lParam;
			pAD = (ad *)((cs)->lpCreateParams);
			SetWindowLong(hwnd, GWL_USERDATA, (LONG)pAD);

			HWND LB_SAMPLERATE = CreateWindow(WC_LISTBOX, "", WS_CHILD | WS_VISIBLE, 50, 10, 100, 310, hwnd, (HMENU)1, cs->hInstance, NULL);
			pAD->LB_SAMPLERATE = LB_SAMPLERATE;

			const int nr_rates = 19;
			int samplerates[nr_rates] = {
				2048, 
				4096, 
				8192, 
				11025,
				22050,
				32768,
				44100,
				48000,
				64000,
				65536,
				88200,
				96000,
				128000,
				131072,
				176400,
				192000,
				256000,
				262144,
				384000			
			};

			for(int i=0; i<nr_rates; i++)
			{
				char buffer[32];
				sprintf_s(buffer, (size_t)32, "%d", samplerates[i]);
				SendMessage(LB_SAMPLERATE, LB_ADDSTRING, 0, (LPARAM)buffer);

				if(samplerates[i] == pAD->SamplesPerSec)
				{
					SendMessage(LB_SAMPLERATE, LB_SETCURSEL, (WPARAM)i, 0);				
				}
			}
		}
		break;
	case WM_CLOSE :

		DestroyWindow(hwnd);
		break;

	case WM_DESTROY :

		{
			int iSamplerate = SendMessage(pAD->LB_SAMPLERATE, LB_GETCURSEL, 0, 0);
			char buffer[100];
			SendMessage(pAD->LB_SAMPLERATE, LB_GETTEXT, (WPARAM)iSamplerate, LPARAM(buffer));
			sscanf_s(buffer, "%d", &pAD->SamplesPerSec);

			pAD->WriteConfig();
		}

		EnableWindow(pAD->BUZZWND, true);
		break;
	}
	return DefWindowProc(hwnd, iMsg, wParam, lParam);
}

void ad::EnableConfigDialog()
{
	char szAppName[] = "Samplerate";

	WNDCLASS wcInstr;
	wcInstr.style         = 0;
	wcInstr.lpfnWndProc   = SettingsDialog;
	wcInstr.cbClsExtra    = 0;
	wcInstr.cbWndExtra    = 0;
	wcInstr.hInstance     = dllInstance;
	wcInstr.hIcon         = LoadIcon(dllInstance, szAppName);
	wcInstr.hCursor       = LoadCursor(NULL, IDC_ARROW);
	wcInstr.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wcInstr.lpszMenuName  = szAppName;
	wcInstr.lpszClassName = szAppName;
	RegisterClass(&wcInstr);

	HWND hWndInstr = CreateWindow(szAppName, szAppName, WS_OVERLAPPEDWINDOW | WS_VISIBLE,  0, 0, 210, 360, NULL, NULL, dllInstance,  (void *)this);
}

void ad::Configure()
{
	BUZZWND = GetActiveWindow();
	EnableWindow(BUZZWND, false);

	EnableConfigDialog();
}

int ad::GetTime()
{
	return 0;
}

int ad::GetPlayPos()
{
	return 0;
}

int ad::GetWritePos()
{
	if (!Running)
		return 0;

	return WritePos & ((1 << 23) - 1);
}

bool ad::Enable(bool e)
{
	if (e)
		return Start();
	else
		return Stop();
}

CAudioDriverInfo info = { 
#ifdef _DEBUG
	"Render To Disk Fast (debug build)"
#else
	"Render To Disk Fast"
#endif
};

extern "C"
{
	__declspec(dllexport) CAudioDriver * __cdecl NewAD()
	{
		return new ad;
	}

	__declspec(dllexport) CAudioDriverInfo const * __cdecl GetADInfo()
	{
		return &info;
	}
}









